(* $Id: configuration.ml,v 1.10 2002/09/02 18:54:44 berdine Exp $ *)

open Types

type goal =
  | GGetPkgsPossible of coords
  | GGetPkgsDefinite of coords * Package.IdSet.t
  | GDeliverPkg of Package.id
  | GExplore

module GoalSet = Set.Make (struct type t = goal let compare = compare end)

let goals = ref (GoalSet.singleton GExplore)

let weight_goal = function
| GGetPkgsPossible(_) -> 2
| GGetPkgsDefinite(_) -> 1
| GDeliverPkg(_) -> 0
| GExplore -> 3

let goal_leq g1 g2 = (weight_goal g1) <= (weight_goal g2)


let money = ref 0

type robot_desc = {mutable posn: coords; mutable pkgs: Package.IdSet.t}

let robots = Hashtbl.create 1024
(* - the size is a guess: we don't know anything about the number of
   robots in a game
   - maybe this should be a Map instead? *)


let self () = Hashtbl.find robots !World.player_rep.World.id

let pkgs_to_deliver_here() = Package.with_dest (self()).pkgs (self()).posn


(* all of these parsing functions could be faster if Str was eliminated, it
   builds intermediate structures when we could just look at the strings *)

let pr_srvr_resp line = print_endline ("> " ^ line)

(* this could be faster if we cached the last lookup *)
let parse_actions r_id =
  let rec loop = function
  | [] -> ()
  | "E" as dir :: tl ->
	  let r_desc = Hashtbl.find robots r_id in
	  begin
		r_desc.posn <- {r_desc.posn with x= r_desc.posn.x + 1};
		loop tl
	  end
  | "W" as dir :: tl ->
	  let r_desc = Hashtbl.find robots r_id in
	  begin
		r_desc.posn <- {r_desc.posn with x= r_desc.posn.x - 1};
		loop tl
	  end
  | "N" as dir :: tl ->
	  let r_desc = Hashtbl.find robots r_id in
	  begin
		r_desc.posn <- {r_desc.posn with y= r_desc.posn.y + 1};
		loop tl
	  end
  | "S" as dir :: tl ->
	  let r_desc = Hashtbl.find robots r_id in
	  begin
		r_desc.posn <- {r_desc.posn with y= r_desc.posn.y - 1};
		loop tl
	  end
  | "P" :: id_str :: tl ->
	  let r_desc = Hashtbl.find robots r_id in
	  let pkg = int_of_string id_str in
	  if r_desc = (self ())
	  then
		begin
		  goals := GoalSet.add (GDeliverPkg pkg) !goals;
		  
(* 	  let pred = function *)
(* 	  | GGetPkgsDefinite(p,_) when p = r_desc.posn -> true *)
(* 	  | _ -> false *)
(* 	  in *)
(* 	  let gs = GoalSet.filter pred !goals  *)
(* 	  begin *)
		  
(* 		(match Hashtbl.find pkg_tiles r_desc.posn with *)
(* 		 | PTDefinite(ps) ->  *)
(* 			 let new_ps = Package.IdSet.remove pkg ps in *)
(* 			 if Package.IdSet.is_empty new_ps *)
(* 			 then Hashtbl.replace pkg_tiles r_desc.posn PTPossible *)
(* 			 else Hashtbl.replace pkg_tiles r_desc.posn (PTDefinite new_ps) *)
(* 		 | _ -> ()); *)

		  r_desc.pkgs <- Package.IdSet.add pkg r_desc.pkgs;
		  loop tl
		end
  | "D" :: id_str :: tl ->
	  let r_desc = Hashtbl.find robots r_id in
	  let pkg = int_of_string id_str in
	  (* TODO: add addition of a PTPossible pkg_tile *)
	  begin
		goals := GoalSet.remove (GDeliverPkg pkg) !goals;
		r_desc.pkgs <- Package.IdSet.remove pkg r_desc.pkgs; loop tl
	  end
  | "X" :: x_str :: "Y" :: y_str :: tl ->
	  let x = int_of_string x_str in
	  let y = int_of_string y_str in
	  let r_desc = {posn= {x= x-1; y= y-1}; pkgs= Package.IdSet.empty} in
	  begin Hashtbl.add robots r_id r_desc; loop tl end
  | _ -> failwith "Invalid action."
  in loop

let parse_update u =
  match Str.split (Str.regexp " ") u with
  | [] -> failwith "Empty update."
  | "Robot" :: id_str :: "died." :: _
	  when (int_of_string id_str) = !World.player_rep.World.id ->
	  failwith "We died."
  | id_str :: action_tokens ->
	  let r_id =
		try int_of_string id_str
		with Failure "int_of_string" ->
		  failwith "Update did not start with an integer."
	  in
	  parse_actions r_id action_tokens

let input_update ichan =
  let line = input_line ichan in
  let updates = Str.split (Str.regexp " *#") line in
  pr_srvr_resp line;
  List.iter parse_update updates


(* this would be faster if we accumulated the multiple changes to one
   mapping together, but it probably doesn't matter *)
let rec parse_packages = function
| [] -> ()
| id_str :: x_str :: y_str :: w_str :: tl ->
	let id = int_of_string id_str in
	let x = (int_of_string x_str) - 1 in
	let y = (int_of_string y_str) - 1 in
	let w = int_of_string w_str in
	let pos = (self()).posn in
	begin
	  Package.new_known id w {x= x; y= y};
	  let pred = function
	  | GGetPkgsPossible(p)
	  | GGetPkgsDefinite(p,_) -> p = pos
	  | _ -> false
	  in
	  let gs = GoalSet.filter pred !goals in
	  let new_gs = ref !goals in
	  let iterer g = match g with
	  | GGetPkgsPossible(c) ->
		  new_gs := (GoalSet.add
					   (GGetPkgsDefinite(c, Package.IdSet.singleton id))
					   (GoalSet.remove g !new_gs))
	  | GGetPkgsDefinite(c,ps) ->
		  new_gs := (GoalSet.add
					   (GGetPkgsDefinite(c, Package.IdSet.add id ps))
					   (GoalSet.remove g !new_gs))
	  | _ -> ()
	  in
	  if GoalSet.is_empty gs
	  then
		goals := (GoalSet.add
					(GGetPkgsDefinite(pos, Package.IdSet.singleton id))
					!goals)
	  else
		begin
		  GoalSet.iter iterer !goals;
		  goals := !new_gs
		end;
	  parse_packages tl
	end
| _ -> failwith "Bad package list."

let input_packages ichan =
  let line = input_line ichan in
  pr_srvr_resp line;
  parse_packages (Str.split (Str.regexp " ") line)


let pr_config () =
  let h = !World.map_rep.World.height in
  let a = Array.make h "" in
  begin
	for j = 0 to h-1 do
	  a.(j) <- String.copy !World.map_rep.World.tiles.(j)
	done;
	let sp = (self ()).posn in a.(sp.y).[sp.x] <- '*';
	for j = 0 to h-1 do
	  print_endline a.(j)
	done;
  end


let pr_goal g =
  print_string "Goal: ";
  match g with
  | GGetPkgsPossible({x= x; y= y}) ->
	  print_endline ("GGetPkgsPossible(" ^ (string_of_int x) ^ " " ^
					 (string_of_int y) ^ ")")
  | GGetPkgsDefinite({x= x; y= y},_) ->
	  print_endline ("GGetPkgsDefinite(" ^ (string_of_int x) ^ " " ^
					 (string_of_int y) ^ ")")
  | GDeliverPkg(p) ->
	  let module P = Package in
	  (match P.desc p with
	   | {P.w=_; P.d= {x= x; y= y}} ->
		   print_endline ("GDeliverPkg(" ^ (string_of_int x) ^ " " ^
						  (string_of_int y) ^ ")"))
  | GExplore -> print_endline "GExplore"
