(* $Id: distance_rects.ml,v 1.7 2002/09/02 18:50:13 richard Exp $ *)

open Types
open World

let get_pos () = (Configuration.self ()).Configuration.posn

(* can't do any longer without compose operator *)

let (<*>) f g x = f (g x)

(* not cat-eliminated ... I'm too lazy today *)
let string_of_int_w i w =
  let s = string_of_int i in
  try String.make (w-String.length s) ' ' ^ s with _ -> s
let string_of_pair fa fb sep (a,b) = "(" ^ fa a ^ sep ^ fb b ^")"
let string_of_triple fa fb fc sep (a,b,c) = "(" ^ fa a ^ sep ^ fb b ^ sep ^ fc c ^ ")"
let string_of_quad fa fb fc fd sep (a,b,c,d) = "(" ^ fa a ^ sep ^ fb b ^ sep ^ fc c ^ sep ^ fd d ^ ")"
let string_of_list f sep xs = "[" ^ String.concat sep (List.map f xs) ^ "]"
let string_of_opt f o =
    match o with
      None -> "None"
    | Some v -> "Some " ^ f v

let show_map {height=height; width=width; tiles=m} =
    let rec pv w k i =
        if k=0 then ()
        else (
            pv w (k/10) (i*10);
            print_string (String.make w ' ');
            for j=1 to width do
                print_string (string_of_int ((j/i) mod 10))
            done;
            print_newline()
        )
    in
    pv 4 width 1;
    for i = height downto 1 do
        print_endline (string_of_int_w i 3 ^ " " ^ m.(i-1))
    done;
    pv 4 width 1
    
(* *****************************************************************

   rectangle abstraction starts here.
   
   Important functions: rects_of_map : map -> node list 
                            (makes the abstraction)
                        shortest_path : (x*y) -> (x'*y') -> node list -> (length * ((node*x*y) list))
                            (tells you how far you have to walk, and how to do it; nodes are included in the path
                             for debugging purposes, but may also be useful if you want to do some local reasoning --
                             e.g. the path tells me to go south, but the second leg is west, so I could step one place
                             west, provided that a west move stays within the node rectangle.
                            )
   
   On a Mac 733Mhz G4 (maybe it's as fast as a 1.4GHz Pentium?), ocaml (top level, presumably not compiled)
   registers shortest_path calculations at about 6/1000 of a second, in  the worst case I can find
   (in the maze; top left to bottom left) -- so you can afford to do a lot of them.
   
   RB 2002/9/1/23/54
   
   *****************************************************************
 *)
 
type pos = int*int (* x,y *)

let internal_pos (x,y) = (pred x, pred y)
let external_pos (x,y) = (succ x, succ y)

let pos_of_coords {x=x;y=y} = (x,y)

type rect = int*int*int*int (* x,x',y,y' -- corner dots of the rectangle *)

let internal_rect (x,x',y,y') = (pred x, pred x', pred y, pred y')
let external_rect (x,x',y,y') = (succ x, succ x', succ y)

type node = {nodenum:int; mutable rect:rect; mutable links:hvlink list; mutable step:step option} 
            (* including node number for ease of printing cyclic structures, dist for running shortest path search *)
and hvlink = Hlink of node | Vlink of node
and step = {pathlength:int; from:int*int*node (* x,y,s *)}

let nodenum = ref 0
let new_node r =
    let i = !nodenum in incr nodenum;
    {nodenum=i;rect=r;links=[];step=None}

let string_of_coord = string_of_int <*> succ
let string_of_pos = string_of_pair string_of_coord string_of_coord ","
let string_of_rect = string_of_quad string_of_coord string_of_coord string_of_coord string_of_coord ","
let string_of_nodenum {nodenum=i} = string_of_int i
let string_of_hvlink hv =
    match hv with
      Hlink h -> "H" ^ string_of_nodenum h
    | Vlink v -> "V" ^ string_of_nodenum v
let string_of_step {pathlength=p; from=x,y,s} = 
    "{" ^ string_of_int p ^ " from " ^ string_of_coord x ^ "," ^ string_of_coord y ^ "(" ^ string_of_nodenum s ^ ")}"
let string_of_node ({rect=r; links=hvs; step=s} as n) = 
    string_of_nodenum n ^ ": " ^ 
    "{" ^ string_of_rect r ^ ";" ^ string_of_list string_of_hvlink "," hvs ^ ";" ^ string_of_opt string_of_step s ^ "}" 
 
let rect_insert_tracing = ref false

(* The efficiency of this function would be increased if I could 
   order the node list so that it can easily
   be combined with new rectangles.  I haven't managed that yet.
 *)
let rec rect_insert (x,x',y,y' as rnew) rs =
    match rs with
      [] ->  (
          let n = new_node rnew in 
          if !rect_insert_tracing then print_endline ("creating node " ^ string_of_node n);
          (Some n,[n])
      )
    | ({rect = rx,rx',ry,ry'} as r)::rs' ->
        let default() = let (n,rs') = rect_insert rnew rs' in (n,r::rs') in
        let project x x' = x,x',min y ry,max y' ry' in
        let hvlink n r rs =
            match n,r with
              None,_ -> n,rs
            | Some ({links=nhvs} as n'),{links=rhvs} -> 
                if !rect_insert_tracing then print_endline (string_of_node n' ^ " HV " ^ string_of_node r);
                n'.links<-Hlink r::nhvs;
                r.links<-Vlink n'::rhvs;
                n,rs
        in
        let vhlink n r rs =
            match n,r with
              None,_ -> n,rs
            | Some ({links=nhvs} as n'),{links=rhvs} -> 
                if !rect_insert_tracing then print_endline (string_of_node n' ^ " VH " ^ string_of_node r);
                n'.links<-Vlink r::nhvs;
                r.links<-Hlink n'::rhvs;
                n,rs
        in
        let rec node_extend (x,x',y,y' as rnew) rs = 
            match rs with
              [] -> ()
            | ({rect = rx,rx',ry,ry'} as r)::rs' ->
                if x<=rx && rx'<=x' && y<=ry' && ry'<=y' then (
                    if !rect_insert_tracing then print_string ("extending " ^ string_of_node r);
                    r.rect<-rx,rx',ry,y';
                    if !rect_insert_tracing then print_endline (" to " ^ string_of_rect r.rect);
                )
                else ();
                node_extend rnew rs'
        in
        if x'<rx then (
            (* r is to right *)
            default() (* this might be a cuttoff ... *) ) else
        if rx'<x then (
            (* r is to left *)
            default() ) else
        if ry'+1<y || y'+1<ry then (
            (* r and rnew don't touch or overlap *)
            default() ) else
        if x=rx && x'=rx' then (
           (* coincidence -- extend r and everything inside it which reaches its southern edge, but no new node *)
           let rr = project x x' in
           if !rect_insert_tracing then print_endline ("combine " ^ string_of_rect rnew ^ "+" ^ string_of_node r ^ "=" ^ string_of_rect rr);
           r.rect<-rr; node_extend rr rs';
           (None, rs) ) else
        if x<=rx && rx'<=x' then (
           (* r is within rnew -- project r across rnew *)
           if !rect_insert_tracing then print_string ("project " ^ string_of_node r ^ " through " ^ string_of_rect rnew ^ "="); 
           r.rect<-project rx rx';
           if !rect_insert_tracing then print_endline (string_of_node r);
           let (n,rs') = rect_insert rnew rs' in
           hvlink n r (r::rs') ) else
        if rx<=x && x'<=rx' then (
           (* rnew is within r -- project rnew across r *)
           let rr = project x x' in
           if !rect_insert_tracing then print_endline ("project " ^ string_of_rect rnew ^ " through " ^ string_of_node r ^ "=" ^ string_of_rect rr);
           let n,rs' = rect_insert rr rs' in
           vhlink n r (r::rs')) else
        if x<=rx then (
           (* overlap rx..x' -- new rectangle *)
           let rr = project rx x' in
           if !rect_insert_tracing then print_endline ("right overlap " ^ string_of_rect rnew ^ " and " ^ string_of_node r ^ "=>" ^ string_of_rect rr);
           let n,rs = default() in
           let _,rs = rect_insert rr rs in
           n,rs ) 
        else (
           (* overlap x..rx' -- new rectangle *)
           let rr = project x rx' in 
           if !rect_insert_tracing then print_endline ("left overlap " ^ string_of_rect rnew ^ " and " ^ string_of_node r ^ "=" ^ string_of_rect rr);
           let n,rs = default() in
           let _,rs = rect_insert rr rs in
           n,rs
        )

let rects_of_map {height=height; tiles=m} =
    let rec rm i nears fars =
        if i=height then nears@fars else (
            let s = m.(i) in
            let len = String.length s in
            let rec rects x =
                if x=len then [] else
                match s.[x] with
                  '.' |'@' ->
                    (let rec r x' =
                         if x'=len then x' else
                         match s.[x'] with
                           '.' |'@' -> r (x'+1)
                         | _ -> x'
                     in
                     let x' = r (x+1) in
                     (x,x'-1,i,i)::rects x' 
                    )
                | _ -> rects (x+1)
            in 
            let nears = List.fold_left (fun rs r -> snd (rect_insert r rs)) nears (rects 0) in
            let nears' = List.filter (fun {rect=_,_,_,y'} -> y'=i) nears in
            let fars' = List.filter (fun {rect=_,_,_,y'} -> y'<>i) nears @ fars in
            rm (i+1) nears' fars'
        )
    in
    nodenum:=0;
    rm 0 [] []

let show_rects rs =
    let rs = List.stable_sort (fun {nodenum=i} {nodenum=j} -> compare i j) rs in
    print_endline (string_of_int (List.length rs) ^ "\n" ^ string_of_list string_of_node ";\n " rs)
    
(* we need a priority queue -- I use a heap a la heapsort, with each element a node and a found distance.
   If the distance doesn't match the node's marked distance, the entry has been superseded.
 *)
exception QueueBomb of string

let new_queue n = Array.make n (new_node (0,0,0,0), 0,0, 0)
let queued_distance (_,_,_,(d:int)) = d

let (<<) a b = queued_distance a < queued_distance b

let rec bubbleup a q n = 
    let parent = n/2 in
    let b = q.(parent) in
    if a<<b then (q.(n)<-b; bubbleup a q parent ) 
    else q.(n)<-a

let qinsert a q n =
    let q = try q.(n+1)<-a; q
            with Invalid_argument _ ->
               (print_endline ("doubling size of queue to "^string_of_int (Array.length q*2));
                Array.append q (new_queue (Array.length q))
               )
    in
    q.(0)<-a; bubbleup a q (n+1); (q,n+1)

(* oh, this is inelegant *)
let rec bubbledown a q k n =
    let child = k*2 in
    let child = if child+1<=n && q.(child+1)<<q.(child) then child+1 else child in
    if child<=n && q.(child)<<a then 
        (q.(k)<-q.(child); bubbledown a q child n) 
    else q.(k)<-a
     
let qremove q n =
    if n=0 then raise (QueueBomb "removing from empty queue") else
    let res = q.(1) in
    bubbledown (q.(n)) q 1 (n-1);
    (res,n-1)

exception PathFail of string

let within (x,y) (rx,rx',ry,ry') = rx<=x && x<=rx' && ry<=y && y<=ry'

let rec unmark node =
  match node with 
    {step=Some _; links=links} ->
      node.step<-None;
      List.iter (function Hlink h -> unmark h | Vlink v -> unmark v) links
  | _ -> ()

let shortest_path_tracing = ref false

let shortest_path (sx,sy as s) (dx,dy as d) rs =
    let sn = try List.find (fun {rect=r} -> within s r) rs with Not_found -> raise (PathFail "invalid source point") in
    let dn = try List.find (fun {rect=r} -> within d r) rs with Not_found -> raise (PathFail "invalid destination point") in
    let rec sp q n =
        if n=0 then raise (PathFail ("no path from "^string_of_pos s^" to "^string_of_pos d))
        else (
            let (s,x,y,len),n = qremove q n in
            match s with 
              {step=Some{pathlength=p}; links=links} when p=len -> (
                if !shortest_path_tracing then 
                    print_endline ("expanding "^string_of_nodenum s^" "^string_of_pos (x,y)^" length "^string_of_int len);
                if within d s.rect then (
                     (* we have a shortest path -- return it *)
                     if !shortest_path_tracing then print_endline ("gotcha!");
                     let rec trace step =
                         match step with 
                           Some {pathlength=p; from=x,y,s} ->
                             if p=0 then [] else (s,x,y)::trace s.step
                         | None -> raise (PathFail "found None in tracing steps")
                     in
                     let rec fulltrace dx dy =
                         if dx<>x then (s,dx,dy)::fulltrace x dy else
                         if dy<>y then (s,dx,dy)::fulltrace dx y else
                         (s,x,y)::trace s.step
                     in
                     len, List.rev (fulltrace dx dy)
                )
                else (
                    (* we have a valid node: expand it *)
                    let rec expand links q n =
                        match links with 
                          l::links -> (
                            let s',len',x',y',step' = 
                                match l with 
                                  Hlink ({rect=rx,rx',ry,ry'; step=hstep} as h) -> 
                                    let xstep = if x<rx then rx-x else if rx'<x then rx'-x else 0 in
                                    h, len+abs xstep, x+xstep, y, hstep
                                | Vlink ({rect=rx,rx',ry,ry'; step=vstep} as v) ->
                                    let ystep = if y<ry then ry-y else if ry'<y then ry'-y else 0 in
                                    v, len+abs ystep, x, y+ystep, vstep
                            in
                            let len' = len' +
                                (if within d s'.rect then (* add the local move to destination to the length *) (
                                     if !shortest_path_tracing then print_string ("found a hit at "^string_of_nodenum s'^"! ");
                                     abs (x'-dx) + abs (y'-dy)
                                 ) else 0)
                            in
                            if (match step' with None->true | Some{pathlength=p}->p>len') then (
                                if !shortest_path_tracing then print_endline ("to "^string_of_nodenum s'^" "^string_of_pos (x',y')^" length "^string_of_int len');
                                s'.step<-Some{pathlength=len';from=x,y,s};
                                let q,n = qinsert (s',x',y',len') q n in
                                expand links q n
                            )
                            else (
                                if !shortest_path_tracing then print_endline ("(forget it -- "^string_of_nodenum s'^" length "^string_of_int len'^")");
                                expand links q n
                            )
                          )
                        | [] -> sp q n
                    in expand links q n
                )
              )
            | _ -> (
                 if !shortest_path_tracing then 
                     print_endline ("abandoning "^string_of_node s^" (length "^string_of_int len^")");
                 sp q n
              )
        )
    in
    sn.step<-Some{pathlength=0;from=sx,sy,sn};
    let q,n = qinsert (sn,sx,sy,0) (new_queue 50 (* small, to see what happens *)) 0 in
    let path = sp q n in
    unmark sn; path

let external_shortest_path s d =
    shortest_path (internal_pos s) (internal_pos d)

let rec show_path (len, ps) = 
   print_endline (string_of_int len ^ " dots; " ^ string_of_int (List.length ps -1) ^ " turns\n" ^
                  string_of_list (string_of_triple string_of_nodenum string_of_coord string_of_coord ", ") "\n "
                                 ps)

let stored_rects = ref None

let get_rects () =
  match !stored_rects with 
    Some rs -> rs
  | None ->
      let rs = rects_of_map !World.map_rep in
      stored_rects := Some rs;
      rs


(* output to Cristiano's interface *)

type dir = {dn:bool; de:bool; ds:bool; dw:bool}

let debug_paths = false

let my_get_path d =
  try
    if debug_paths then print_endline ("path from "^string_of_pos (pos_of_coords (get_pos()))^" to "^string_of_pos (pos_of_coords d));
    let r = shortest_path (pos_of_coords (get_pos())) (pos_of_coords d) (get_rects()) in
    if debug_paths && List.length (snd r)<4 then show_path r;
    Some r
  with PathFail "no path" -> None

let xy_of_sxy (s,x,y) = x,y

let get_path d =
  match my_get_path d with
    Some (len,xys) ->
      Some (List.map xy_of_sxy (List.tl xys), len)
  | None -> None

let single dx = if dx=0 then 0 else dx/(abs dx) (* 0, +1 or -1 *)

let get_dirs d =
  match my_get_path d with
    Some(len,sxys) ->
        let rec getd sxys = 
		  match sxys with 
			[_] -> if debug_paths then print_endline ("no step possible"); 
			       {dn=false; de=false; ds=false; dw=false} (* no step needed *)
		  | _ ->
			  let (_x,x0,y0),(_,x1,y1) = List.hd sxys, List.hd(List.tl sxys) in
			  let dx,dy = single(x1-x0), single(y1-y0) in
			  if dx=0 && dy=0 then getd (List.tl sxys) else 
				let rec deviation sxys =
				  match sxys with
					((s0,x0,y0) as p0)::((s1,x1,y1)::(s2,x2,y2)::_ as sxys) ->
					  (let test =
						 function None              -> None
						 |        Some (dx,dy) as r -> if within (x0+dx,y0+dy) s0.rect then r else None
					   in
					   match single (x1-x0), single(y1-y0), single (x2-x1), single (y2-y1) with
							 0             , _            , 0             , _     -> test (deviation sxys)
					   |     _             , 0            , _             , 0     -> test (deviation sxys)
					   |     dx1           , dy1          , dx2           , dy2   -> test (Some(dx2,dy2)))
				  | _ -> None
				in
				let dx',dy'= 
				   match deviation sxys with
					 None -> dx,dy
				   | Some(dx1,dy1) -> dx+dx1,dy+dy1
				in
				if debug_paths then print_endline ((if dy'>0 then "N " else "")^(if dx'>0 then "E " else "")^
							                      (if dy'<0 then "S " else "")^(if dx'<0 then "W " else ""));
				{dn=(dy'>0); de=(dx'>0); ds=(dy'<0); dw=(dx'<0)}
        in
        Some(getd sxys, len)
  | None -> None
  
